#!/usr/bin/python
# -*- coding:utf-8 -*-

#------------------------------------------------------------------------------
#   SOMA PUZZLE SOLVER
#   Author :   monnomamoi
#   Contact:   <http://www.developpez.net/forums/u138273/monnomamoi/>
#   Licence:   <http://creativecommons.org/publicdomain/zero/1.0/>
#------------------------------------------------------------------------------

"""Solveur de puzzles 3D utilisant les pièces du cube Soma.

Utilisation en ligne de commande :
    ./puzzle.py  [-i fichier_de_points]  [-o fichier_de_sortie]
    ./puzzle.py  -h
    ./puzzle.py  -v
Par défaut, les points sont ceux d'un cube 3x3x3, et la sortie est 'stdout'."""

__version__ = '0.4.1'
__date__    = '2009-10-08'


import sys
import getopt



#===# FONCTIONS #==============================================================


def read_file(filename):
    """Charge une liste de points depuis un fichier."""
    data = []
    try:
        for line in open(filename, 'r'):
            pt = map(int, line.split(','))
            if len(pt) != 3:
                raise ValueError
            data.append(tuple(pt))
    except:
        data = None
    return data


def write_file(filename, data):
    """Enregistre une liste de points dans un fichier."""
    points = [pt+(i+1,) for i,pc in enumerate(data) for pt in pc[1]]
    try:
        f = open(filename, 'w')
        f.writelines(','.join(map(str, pt))+'\n' for pt in points)
        f.close()
        return True
    except:
        return False


def gen_basis():
    """Génère la liste des 24 matrices de changement de base."""
    z = [0, 0, 0]
    # Calcule toutes les combinaisons possibles des 2 premières lignes
    mat = [ [z[:l]+[k]+z[l+1:], z[:j]+[i]+z[j+1:]]  for i in [1,-1]
                                                    for j in xrange(3)
                                                    for k in [1,-1]
                                                    for l in xrange(3)
                                                    if j!=l ]
    # Base orthonormale directe : ligne[3] = ligne[1] ^ ligne[2]
    for m in mat:
        m.append( [ m[0][1]*m[1][2] - m[0][2]*m[1][1],
                    m[0][2]*m[1][0] - m[0][0]*m[1][2],
                    m[0][0]*m[1][1] - m[0][1]*m[1][0] ] )
    return mat



#===# CLASSES #================================================================


class Shape:
    """Forme (vide) d'un puzzle."""
    
    def __init__(self, points):
        self.points = set(points)
    
    def check(self):
        """Vérifie la consistance des données."""
        cond = [len(self.points) > 0]                                           # teste si ensemble non vide
        cond.append(min(map(min, self.points)) >= 0)                            # teste si coordonnées >= 0
        cond.extend(len(pt) == 3 for pt in self.points)                         # teste si 3 dimensions
        return all(cond)



#------------------------------------------------------------------------------


class FillableShape(Shape):
    """Forme d'un puzzle, pouvant être remplie."""
    
    def __init__(self, points):
        Shape.__init__(self, points)
        self.pieces = []
    
    def list_used_pieces(self):
        """Liste les UID des pièces utilisées."""
        return (pc[0] for pc in self.pieces)
    
    def push_piece(self, uid, pts):
        """Insère une pièce ; renvoie False si impossible."""
        if not pts.issubset(self.points):
            return False
        self.pieces.append((uid, pts))
        self.points.difference_update(pts)
        return True
    
    def pop_piece(self):
        """Enlève la dernière pièce ajoutée."""
        self.points.update(self.pieces[-1][1])
        self.pieces.pop()
        return



#------------------------------------------------------------------------------


class Piece(Shape):
    """Pièce pour puzzle."""
    
    def __init__(self, uid, points):
        Shape.__init__(self, points)
        self.uid = uid
        self.configs = []
        self.__gen_configs()
    
    def translate(self, c, pt, pos):
        """Translate point 'pt' de config 'c' vers position 'pos'."""
        return set(tuple(p[i]-pt[i]+pos[i] for i in xrange(3)) for p in c)
    
    def __gen_configs(self):
        """Génère l'ensemble des configurations spatiales possibles."""
        for b in gen_basis():
            # Changement de base :  X' = B*X
            c = [tuple(sum(p[i]*b[j][i] for i in xrange(3))
                                        for j in xrange(3))
                                        for p in self.points]
            m = [min(p[i] for p in c) for i in xrange(3)]                       # calcul des minima
            c = set(tuple(p[i]-m[i] for i in xrange(3)) for p in c)             # recentrer
            if not any(map(c.issubset, self.configs)):
                self.configs.append(c)                                          # ajout si pas de doublon
        return



#------------------------------------------------------------------------------


class Puzzle:
    """Puzzle 3D défini par sa forme finale et ses pièces."""
    
    def __init__(self, shape, pieces):
        self.shape = shape
        self.pieces = list(pieces)
        self.minspace = min(len(pc.points) for pc in pieces)
    
    def check(self):
        """Vérifie la consistance des données."""
        cond = [self.shape.check()] + [pc.check() for pc in self.pieces]
        s = sum(len(pc.points) for pc in self.pieces)
        cond.append(s == len(self.shape.points))                                # teste si même nombre de points
        return all(cond)
    
    def solve(self):
        """Résoud le puzzle."""
        f = FillableShape(set(self.shape.points))
        r = self.__solve_rec(f)
        return f.pieces if r else None
    
    def __fit(self, dim, f):
        """Vérifie si les pièces libres de 'f' peuvent remplir 'dim' trous."""
        lpcs = [ len(pc.points) for pc in self.pieces
                                if pc.uid not in f.list_used_pieces() ]
        l = list(set((i,lpcs.count(i)+1) for i in lpcs))
        total = reduce(int.__mul__, (n[1] for n in l))
        # Résolution de l'équation diophantienne correspondante
        for k in xrange(total):
            i,j = 0,k
            for n,c in l:
                j,m = divmod(j,c)
                i += n*m
            if i == dim:
                return True
        return False
    
    def __prune(self, points):
        """Ne garde que le plus petit ensemble contigu de points."""
        pts = set(points)
        zones = []                                                              # initialise l'ensemble des zones disjointes
        while len(pts) > 0:
            zones.append(set([pts.pop()]))                                      # point libre -> nouvelle zone
            n = True
            while n:                                                            # ajoute points adjacents tant qu'il y en a
                n = set( pt for pt in pts for z in zones[-1]
                            if sum(abs(pt[i]-z[i]) for i in xrange(3)) == 1 )
                zones[-1].update(n)
                pts.difference_update(n)
        m = map(len, zones)                                                     # calcule taille des zones
        return zones[m.index(min(m))]                                           # retourne la plus petite
    
    def __solve_rec(self, f):
        """Résolution récursive du puzzle."""
        minizone = self.__prune(f.points)                                       # limite les points à tester
        if len(minizone) < self.minspace:                                       # stoppe si espace trop petit
            return False
        if not self.__fit(len(minizone), f):                                    # stoppe si pas remplissable
            return False
        x = minizone.pop()                                                      # choisis un point dans minizone
        for pc in self.pieces:                                                  # parcourt les pièces
            if pc.uid not in f.list_used_pieces():
                for c in pc.configs:                                            # parcourt les configurations
                    for k in c:                                                 # parcourt les points de la config
                        t = pc.translate(c, k, x)
                        if f.push_piece(pc.uid, t):                             # si la pièce rentre
                            ###print list(f.list_used_pieces())                 # LOG
                            if len(f.points) == 0:                              # fin si complet
                                return True
                            elif self.__solve_rec(f):                           # sinon, continue
                                return True
                            else:
                                f.pop_piece()                                   # retire pièce si impossible
        return False



#------------------------------------------------------------------------------


class Soma(Puzzle):
    """Puzzle dont les pièces sont celles du cube Soma."""
    
    def __init__(self, shape):
        pcs = [ Piece("V", [(0,0,0), (1,0,0), (0,1,0)]),
                Piece("L", [(0,0,0), (1,0,0), (2,0,0), (0,1,0)]),
                Piece("T", [(0,0,0), (1,0,0), (2,0,0), (1,1,0)]),
                Piece("Z", [(0,0,0), (1,0,0), (1,1,0), (2,1,0)]),
                Piece("A", [(0,0,0), (0,1,0), (0,0,1), (1,0,1)]),
                Piece("B", [(0,0,0), (0,1,0), (1,0,0), (1,0,1)]),
                Piece("P", [(0,0,0), (1,0,0), (0,1,0), (0,0,1)]) ]
        Puzzle.__init__(self, shape, pcs)



#------------------------------------------------------------------------------


class Soma3x3x3(Soma):
    """Puzzle correspondant au cube Soma 3x3x3."""
    
    def __init__(self):
        pts=[(i,j,k) for i in xrange(3) for j in xrange(3) for k in xrange(3)]
        Soma.__init__(self, Shape(pts))



#===# MAIN #===================================================================


if __name__ == '__main__':
    
    # Arguments de la ligne de commande #
    try:
        options = dict(getopt.getopt(sys.argv[1:], "i:o:hv")[0])
    except:
        print "ERREUR : mauvais arguments !"
        print "Afficher l'aide :   ./puzzle.py -h"
        sys.exit(2)
    if "-h" in options:
        print __doc__
        sys.exit(0)
    if "-v" in options:
        print "Version : %s" % (__version__)
        sys.exit(0)
    
    # Source des données #
    if "-i" in options:
        f = options["-i"]
        print "Fichier d'entrée : %s" % (f)
        pts = read_file(f)
        if not pts:
            print "ERREUR : fichier inaccessible ou mal construit !"
            sys.exit(1)
        pz = Soma(Shape(pts))
        if max(map(max, pts)) > 9 or not pz.check():
            print "ERREUR : données inconsistantes !"
            sys.exit(1)
    else:
        print "Cube de Soma 3x3x3"
        pz = Soma3x3x3()
    
    # Résolution #
    s = pz.solve()
    
    # Restitution des résultats #
    if s:
        print "Solution trouvée   :)"
        if "-o" in options:
            f = options["-o"]
            print "Fichier de sortie : %s" % (f)
            if not write_file(f, s):
                print "ERREUR : écriture impossible !"
                sys.exit(1)
        else:
            print "---"
            for pc in s:
                print "Pièce %s : %s" % (str(pc[0]), str(list(pc[1]))[1:-1])
    else:
        print "Pas de solution   :("



#==============================================================================
